# "Export" most used elements
from . import evaluator, interpreter, io, model, stories, testing

__description__ = 'Sismic Interactive State Machine Interpreter and Checker'
__version__ = '0.14.3'
__url__ = 'https://github.com/AlexandreDecan/sismic/'
__author__ = 'Alexandre Decan'
__email__ = 'alexandre.decan@lexpage.net'
__licence__ = 'LGPL3'